@@ -3,7 +3,7 @@
 /* header file for S-Lang internal structures that users do not (should not)
    need.  Use slang.h for that purpose. */
 /*
-Copyright (C) 2004-2010 John E. Davis
+Copyright (C) 2004-2016 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -484,19 +484,19 @@ typedef unsigned int _pSLuint64_Type;
 # if SIZEOF_SHORT == 8
 #  define _pSLANG_INT64_TYPE	SLANG_SHORT_TYPE
 #  define _pSLANG_UINT64_TYPE	SLANG_USHORT_TYPE
-typedef int _pSLInt64_Type;
+typedef int _pSLint64_Type;
 typedef unsigned int _pSLuint64_Type;
 # else
 #  if SIZEOF_LONG == 8
 #   define _pSLANG_INT64_TYPE	SLANG_LONG_TYPE
 #   define _pSLANG_UINT64_TYPE	SLANG_ULONG_TYPE
-typedef long _pSLInt64_Type;
+typedef long _pSLint64_Type;
 typedef unsigned long _pSLuint64_Type;
 #  else
 #   if SIZEOF_LONG_LONG == 8
 #    define _pSLANG_INT64_TYPE	SLANG_LLONG_TYPE
 #    define _pSLANG_UINT64_TYPE	SLANG_ULLONG_TYPE
-typedef long long _pSLInt64_Type;
+typedef long long _pSLint64_Type;
 typedef unsigned long long _pSLuint64_Type;
 #   else
 #    define _pSLANG_INT64_TYPE	0
@@ -506,6 +506,34 @@ typedef unsigned long long _pSLuint64_Type;
 # endif
 #endif
 
+/* The following are commented out because they are not used by the library */
+/* extern int _pSLang_pop_int16 (_pSLint16_Type *); */
+/* extern int _pSLang_pop_uint16 (_pSLuint16_Type *); */
+/* extern int _pSLang_pop_int32 (_pSLint32_Type *); */
+/* extern int _pSLang_pop_uint32 (_pSLuint32_Type *); */
+#if _pSLANG_INT64_TYPE
+extern int _pSLang_pop_int64(_pSLint64_Type *);
+extern int _pSLang_pop_uint64 (_pSLuint64_Type *);
+#endif
+/* extern int _pSLang_push_int16 (_pSLint16_Type); */
+/* extern int _pSLang_push_uint16 (_pSLuint16_Type); */
+/* extern int _pSLang_push_int32 (_pSLint32_Type); */
+/* extern int _pSLang_push_uint32 (_pSLuint32_Type); */
+#if _pSLANG_INT64_TYPE
+/* extern int _pSLang_push_int64(_pSLint64_Type); */
+/* extern int _pSLang_push_uint64 (_pSLuint64_Type); */
+#endif
+
+#ifdef HAVE_LONG_LONG
+# ifdef __WIN32__
+#  define SLFMT_LLD  "%I64d"
+#  define SLFMT_LLU  "%I64u"
+# else
+#  define SLFMT_LLD  "%lld"
+#  define SLFMT_LLU  "%llu"
+# endif
+#endif
+
 typedef union
 {
 #if SLANG_HAS_FLOAT
@@ -531,6 +559,14 @@ typedef union
    char char_val;
    unsigned char uchar_val;
    SLindex_Type index_val;
+   _pSLint16_Type int16_val;
+   _pSLuint16_Type uint16_val;
+   _pSLint32_Type int32_val;
+   _pSLuint32_Type uint32_val;
+#if _pSLANG_INT64_TYPE
+   _pSLint64_Type int64_val;
+   _pSLuint64_Type uint64_val;
+#endif
 }
 _pSL_Object_Union_Type;
 
@@ -589,11 +625,6 @@ extern int _pSLstruct_pop_field (SLang_Struct_Type *s, SLFUTURE_CONST char *name
 
 extern int _pSLang_get_qualifiers_intrin (SLang_Struct_Type **);
 
-/* The following qualifier functions will eventually be part of the public API */
-extern int _pSLang_qualifier_exists (SLCONST char *);
-extern int _pSLang_get_int_qualifier (SLCONST char *, int *, int);
-extern int _pSLang_get_string_qualifier (SLCONST char *, char **p, SLFUTURE_CONST char *);
-
 struct _pSLang_Ref_Type
 {
    int num_refs;
@@ -635,14 +666,14 @@ extern int _pSLpush_malloced_string (char *, unsigned int);
  * very careful how they are used.  In particular, if len bytes are allocated,
  * then the string must be len characters long, no more and no less.
  */
-extern char *_pSLallocate_slstring (unsigned int);
-extern char *_pSLcreate_via_alloced_slstring (char *, unsigned int);
-extern void _pSLunallocate_slstring (char *, unsigned int);
-extern int _pSLpush_alloced_slstring (char *, unsigned int);
+extern char *_pSLallocate_slstring (size_t);
+extern char *_pSLcreate_via_alloced_slstring (char *, size_t);
+extern void _pSLunallocate_slstring (char *, size_t);
+extern int _pSLpush_alloced_slstring (char *, size_t);
 
-extern unsigned int _pSLstring_bytelen (SLstr_Type *);
+extern size_t _pSLstring_bytelen (SLCONST SLstr_Type *);
 extern void _pSLang_free_slstring (SLstr_Type *);   /* slstring required and assumed */
-extern unsigned long _pSLstring_get_hash (SLstr_Type *s);   /* slstring required */
+extern SLstr_Hash_Type _pSLstring_get_hash (SLstr_Type *s);   /* slstring required */
 
 typedef struct
 {
@@ -703,6 +734,8 @@ struct _pSLang_NameSpace_Type
    unsigned int table_size;
    SLang_Name_Type **table;
 };
+
+extern int _pSLns_check_name (SLFUTURE_CONST char *);
 extern SLang_NameSpace_Type *_pSLns_new_namespace (SLFUTURE_CONST char *, unsigned int);
 extern SLang_NameSpace_Type *_pSLns_allocate_namespace (SLFUTURE_CONST char *, unsigned int);
 extern void _pSLns_deallocate_namespace (SLang_NameSpace_Type *);
@@ -714,8 +747,8 @@ extern SLFUTURE_CONST char *_pSLang_cur_namespace_intrinsic (void);
 extern SLang_Array_Type *_pSLang_apropos (SLFUTURE_CONST char *, SLFUTURE_CONST char *, unsigned int);
 extern void _pSLang_implements_intrinsic (SLFUTURE_CONST char *);
 extern SLang_Array_Type *_pSLns_list_namespaces (void);
-extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, SLCONST char *, unsigned long);
-extern int _pSLns_add_hashed_name (SLang_NameSpace_Type *, SLang_Name_Type *, unsigned long);
+extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, SLCONST char *, SLstr_Hash_Type);
+extern int _pSLns_add_hashed_name (SLang_NameSpace_Type *, SLang_Name_Type *, SLstr_Hash_Type);
 extern SLang_NameSpace_Type *_pSLns_find_object_namespace (SLang_Name_Type *nt);
 extern SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *, SLCONST char *);
 extern SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, SLFUTURE_CONST char *nsname);
@@ -753,6 +786,8 @@ extern int _pSLang_trace_fun(SLFUTURE_CONST char *);
 /* extern int _pSLang_Compile_Line_Num_Info; */
 #if SLANG_HAS_BOSEOS
 extern int _pSLang_Compile_BOSEOS;
+#define SLANG_BOSEOS_VALUE_BITS		0x0FF
+#define SLANG_BOSEOS_PREPROC		0x100
 extern int _pSLang_Compile_BOFEOF;
 extern int _pSLang_init_boseos (void);
 extern int _pSLcall_bos_handler (SLFUTURE_CONST char *, int);
@@ -763,11 +798,10 @@ extern int _pSLcall_debug_hook (SLFUTURE_CONST char *file, int linenum);
 /* extern int _pSLcall_debug_hook (char *file, int linenum, char *funct); */
 #endif
 
-extern char *_pSLstring_dup_hashed_string (SLCONST char *, unsigned long);
-extern unsigned long _pSLcompute_string_hash (SLCONST char *);
-extern char *_pSLstring_make_hashed_string (SLCONST char *, unsigned int, unsigned long *);
-extern void _pSLfree_hashed_string (char *, unsigned int, unsigned long);
-unsigned long _pSLstring_hash (SLCONST unsigned char *, SLCONST unsigned char *);
+extern char *_pSLstring_dup_hashed_string (SLCONST char *, SLstr_Hash_Type);
+extern char *_pSLstring_make_hashed_string (SLCONST char *, SLstrlen_Type, SLstr_Hash_Type *);
+extern void _pSLfree_hashed_string (SLCONST char *, size_t, SLstr_Hash_Type);
+SLstr_Hash_Type _pSLstring_hash (SLCONST unsigned char *, SLCONST unsigned char *);
 extern int _pSLinit_slcomplex (void);
 
 extern int _pSLang_init_slstrops (void);
@@ -784,7 +818,7 @@ extern int _pSLang_init_sltime (void);
 extern void _pSLpack (void);
 extern void _pSLunpack (char *, SLang_BString_Type *);
 extern void _pSLpack_pad_format (char *);
-extern unsigned int _pSLpack_compute_size (char *);
+extern SLstrlen_Type _pSLpack_compute_size (char *);
 extern int _pSLusleep (unsigned long);
 
 /* frees upon error.  NULL __NOT__ ok. */
@@ -957,10 +991,10 @@ struct _pSLang_Class_Type
    int (*cl_sget) (SLtype, SLFUTURE_CONST char *);
 
    /* File I/O */
-   int (*cl_fread) (SLtype, FILE *, VOID_STAR, unsigned int, unsigned int *);
-   int (*cl_fwrite) (SLtype, FILE *, VOID_STAR, unsigned int, unsigned int *);
-   int (*cl_fdread) (SLtype, int, VOID_STAR, unsigned int, unsigned int *);
-   int (*cl_fdwrite) (SLtype, int, VOID_STAR, unsigned int, unsigned int *);
+   int (*cl_fread) (SLtype, FILE *, VOID_STAR, SLstrlen_Type, SLstrlen_Type *);
+   int (*cl_fwrite) (SLtype, FILE *, VOID_STAR, SLstrlen_Type, SLstrlen_Type *);
+   int (*cl_fdread) (SLtype, int, VOID_STAR, SLstrlen_Type, SLstrlen_Type *);
+   int (*cl_fdwrite) (SLtype, int, VOID_STAR, SLstrlen_Type, SLstrlen_Type *);
 
    int (*cl_to_bool) (SLtype, int *);
 
@@ -1069,6 +1103,8 @@ extern SLFUTURE_CONST char *_pSLdefines[];
 #define SL_ERRNO_NOT_IMPLEMENTED	0x7FFF
 extern int _pSLerrno_errno;
 extern int _pSLerrno_init (void);
+extern SLFUTURE_VOID *_SLcalloc (SLstrlen_Type, SLstrlen_Type);
+extern SLFUTURE_VOID *_SLrecalloc (SLFUTURE_VOID *, SLstrlen_Type, SLstrlen_Type);
 
 extern int _pSLstdio_fdopen (char *, int, char *);
 extern void _pSLfclose_fdopen_fp (SLang_MMT_Type *);
@@ -1139,7 +1175,7 @@ typedef struct
 	SLFUTURE_CONST char *s_val;		       /* (SLstring) concatenated strings */
 	SLang_BString_Type *b_val;     /* concatenated bstrings */
      } v;
-   unsigned long hash;		       /* for v.s_val */
+   SLstr_Hash_Type hash;		       /* for v.s_val */
    unsigned int len;		       /* length of v.?_val */
 }
 _pSLang_Multiline_String_Type;
@@ -1188,7 +1224,7 @@ _pSLang_Token_Type;
 
 /* return token type or EOF_TOKEN upon error */
 extern _pSLtok_Type _pSLtoken_init_slstring_token (_pSLang_Token_Type *, _pSLtok_Type,
-						   SLFUTURE_CONST char *, unsigned int);
+						   SLCONST char *, SLstrlen_Type);
 
 extern void _pSLcompile (_pSLang_Token_Type *);
 extern void (*_pSLcompile_ptr)(_pSLang_Token_Type *);
@@ -1608,8 +1644,15 @@ extern int _pSLang_check_signals_hook (VOID_STAR);
 #endif
 
 #undef _INLINE_
-#if defined(__GNUC__) && SLANG_USE_INLINE_CODE
-# define _INLINE_ __inline__
+/* clang defines GNUC on the mac, but used c99 sematics for extern inline.  This implementation
+ * assume GNU semantics.
+ */
+#if defined (SLANG_USE_INLINE_CODE) && defined(__GNUC__) && !defined(__clang__)
+# if !defined __GNUC_STDC_INLINE__ && !defined __GNUC_GNU_INLINE__
+#  define _INLINE_ __inline__    /* older compilers */
+# else
+#  define _INLINE_ __inline__ __attribute__((gnu_inline))    /* newer ones */
+# endif
 #else
 # define _INLINE_
 #endif
