@@ -2,7 +2,7 @@
  * links to the library.
  */
 /*
-Copyright (C) 2004-2010 John E. Davis
+Copyright (C) 2004-2016 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -42,16 +42,10 @@ USA.
 # include <windows.h>
 #endif
 
-#define DEBUG_MALLOC 0
-
-#if DEBUG_MALLOC
-# define SLREALLOC_FUN	SLdebug_realloc
-# define SLMALLOC_FUN	SLdebug_malloc
-# define SLFREE_FUN	SLdebug_free
-#else
-# define SLREALLOC_FUN	SLREALLOC
-# define SLMALLOC_FUN	SLMALLOC
-# define SLFREE_FUN	SLFREE
+#define SLSYSWRAP_REALLOC realloc
+#define SLSYSWRAP_MALLOC malloc
+#ifdef SLSYSWRAP
+# include <slsyswrap.h>
 #endif
 
 int SLang_Version = SLANG_VERSION;
@@ -100,6 +94,14 @@ static int utf8_enable (int mode)
 
 #ifdef HAVE_NL_LANGINFO_CODESET
    locale = nl_langinfo (CODESET);
+# ifdef SLSYSWRAP
+     {
+	char *env;
+	if ((NULL != (env = getenv ("SLTEST_NO_LANGINFO")))
+	    && (atoi (env) != 0))
+	  locale = NULL;		       /* skip this block of code */
+     }
+# endif
    if ((locale != NULL) && (*locale))
      {
 	if ((0 == strcmp (locale, "UTF-8"))
@@ -169,25 +171,25 @@ int SLutf8_enable (int mode)
  * and the library malloc function returns NULL, then malloc will be retried
  * using 1 byte.
  */
-char *SLmalloc (unsigned int len)
+SLFUTURE_VOID *SLmalloc (SLstrlen_Type len)
 {
-   char *p;
+   SLFUTURE_VOID *p;
 
-   if (NULL != (p = (char *) SLMALLOC_FUN (len)))
+   if (NULL != (p = (SLFUTURE_VOID *)SLSYSWRAP_MALLOC(len)))
      return p;
 
-   if (len || (NULL == (p = (char *)SLMALLOC_FUN(1))))
+   if (len || (NULL == (p = (SLFUTURE_VOID *)SLSYSWRAP_MALLOC(1))))
      SLang_set_error (SL_MALLOC_ERROR);
 
    return p;
 }
 
-void SLfree (char *p)
+void SLfree (SLFUTURE_VOID *p)
 {
-   if (p != NULL) SLFREE_FUN (p);
+   if (p != NULL) free (p);
 }
 
-char *SLrealloc (char *p, unsigned int len)
+SLFUTURE_VOID *SLrealloc (SLFUTURE_VOID *p, SLstrlen_Type len)
 {
    if (len == 0)
      {
@@ -198,20 +200,41 @@ char *SLrealloc (char *p, unsigned int len)
    if (p == NULL) p = SLmalloc (len);
    else
      {
-	p = (char *)SLREALLOC_FUN (p, len);
+	p = (SLFUTURE_VOID *) SLSYSWRAP_REALLOC(p, len);
 	if (p == NULL)
 	  SLang_set_error (SL_MALLOC_ERROR);
      }
    return p;
 }
 
-char *SLcalloc (unsigned int nelems, unsigned int len)
+SLFUTURE_VOID *_SLrecalloc (SLFUTURE_VOID *p, SLstrlen_Type nelems, SLstrlen_Type len)
 {
-   char *p;
+   unsigned int nlen = nelems * len;
 
-   len = nelems * len;
-   p = SLmalloc (len);
-   if (p != NULL) memset (p, 0, len);
+   if (nelems && (nlen/nelems != len))
+     {
+	SLang_set_error (SL_Malloc_Error);
+	return NULL;
+     }
+   return SLrealloc (p, nlen);
+}
+
+SLFUTURE_VOID *_SLcalloc (SLstrlen_Type nelems, SLstrlen_Type len)
+{
+   SLstrlen_Type nlen = nelems * len;
+
+   if (nelems && (nlen/nelems != len))
+     {
+	SLang_set_error (SL_Malloc_Error);
+	return NULL;
+     }
+   return SLmalloc (nlen);
+}
+
+SLFUTURE_VOID *SLcalloc (SLstrlen_Type nelems, SLstrlen_Type len)
+{
+   SLFUTURE_VOID *p = _SLcalloc (nelems, len);
+   if (p != NULL) memset (p, 0, len*nelems);
    return p;
 }
 
@@ -269,6 +292,7 @@ static Interrupt_Hook_Type *
 	       *prevp = prev;
 	     return h;
 	  }
+	prev = h;
 	h = h->next;
      }
    return NULL;
@@ -337,6 +361,9 @@ int SLerrno_set_errno (int sys_errno)
 #if defined(__WIN32__) && defined(SLANG_DLL)
 # include <windows.h>
 
+/* where is the prototype for DllMain? */
+BOOL WINAPI DllMain(HANDLE hInstance,DWORD dwReason,LPVOID lpParam);
+
 BOOL WINAPI DllMain(HANDLE hInstance,DWORD dwReason,LPVOID lpParam)
 {
    return 1;
