@@ -1,6 +1,6 @@
 /* List of objects */
 /*
-Copyright (C) 2004-2010 John E. Davis
+Copyright (C) 2004-2016 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -40,27 +40,27 @@ USA.
  *    copy (list)
  */
 
-typedef struct _pSLang_List_Type SLang_List_Type;
-
-#define CHUNK_SIZE 128
-
 typedef struct _Chunk_Type
 {
    struct _Chunk_Type *next;
    struct _Chunk_Type *prev;
-   int num_elements;
-   SLang_Object_Type *elements;	       /* CHUNK_SIZE of em */
+   SLindex_Type num_elements;
+   SLindex_Type chunk_size;
+   SLang_Object_Type *elements;	       /* chunk_size of em */
 }
 Chunk_Type;
 
 struct _pSLang_List_Type
 {
    SLindex_Type length;
+#define DEFAULT_CHUNK_SIZE	128
+   SLuindex_Type default_chunk_size;
    Chunk_Type *first;
    Chunk_Type *last;
 
    Chunk_Type *recent;		       /* most recent chunk accessed */
    SLindex_Type recent_num;	       /* num elements before the recent chunk */
+   int ref_count;
 };
 
 static void delete_chunk (Chunk_Type *c)
@@ -79,7 +79,7 @@ static void delete_chunk (Chunk_Type *c)
    SLfree ((char *) c);
 }
 
-static Chunk_Type *new_chunk (void)
+static Chunk_Type *new_chunk (unsigned int size)
 {
    Chunk_Type *c;
 
@@ -87,12 +87,13 @@ static Chunk_Type *new_chunk (void)
    if (c == NULL)
      return c;
 
-   c->elements = (SLang_Object_Type *)SLcalloc(CHUNK_SIZE, sizeof(SLang_Object_Type));
+   c->elements = (SLang_Object_Type *)SLcalloc(size, sizeof(SLang_Object_Type));
    if (c->elements == NULL)
      {
 	SLfree ((char *) c);
 	return NULL;
      }
+   c->chunk_size = size;
    return c;
 }
 
@@ -107,34 +108,34 @@ static void delete_chunk_chain (Chunk_Type *c)
      }
 }
 
-static void delete_list (SLang_List_Type *list)
+static void free_list (SLang_List_Type *list)
 {
    if (list == NULL)
      return;
 
+   if (list->ref_count > 1)
+     {
+	list->ref_count--;
+	return;
+     }
+
    delete_chunk_chain (list->first);
    SLfree ((char *) list);
 }
 
-static void list_destroy (SLtype type, VOID_STAR ptr)
-{
-   (void) type;
-   delete_list ((SLang_List_Type *) ptr);
-}
-
-static int make_chunk_chain (int length, Chunk_Type **firstp, Chunk_Type **lastp)
+static int make_chunk_chain (SLindex_Type length, Chunk_Type **firstp, Chunk_Type **lastp, int chunk_size)
 {
    Chunk_Type *last, *first;
 
-   if (NULL == (first = new_chunk ()))
+   if (NULL == (first = new_chunk (chunk_size)))
      return -1;
-   length -= CHUNK_SIZE;
+   length -= chunk_size;
 
    last = first;
    while (length > 0)
      {
 	Chunk_Type *next;
-	if (NULL == (next = new_chunk ()))
+	if (NULL == (next = new_chunk (chunk_size)))
 	  {
 	     delete_chunk_chain (first);
 	     return -1;
@@ -142,16 +143,29 @@ static int make_chunk_chain (int length, Chunk_Type **firstp, Chunk_Type **lastp
 	last->next = next;
 	next->prev = last;
 	last = next;
-	length -= CHUNK_SIZE;
+	length -= chunk_size;
      }
    *firstp = first;
    *lastp = last;
    return 0;
 }
 
-static SLang_List_Type *allocate_list (void)
+static SLang_List_Type *allocate_list (SLindex_Type chunk_size)
 {
-   return (SLang_List_Type *)SLcalloc (1, sizeof (SLang_List_Type));
+   SLang_List_Type *list;
+
+   if (chunk_size <= 0)
+     chunk_size = DEFAULT_CHUNK_SIZE;
+   else if (chunk_size > 2*DEFAULT_CHUNK_SIZE)
+     chunk_size = 2*DEFAULT_CHUNK_SIZE;
+
+   list = (SLang_List_Type *)SLcalloc (1, sizeof (SLang_List_Type));
+   if (list != NULL)
+     {
+	list->ref_count = 1;
+	list->default_chunk_size = chunk_size;
+     }
+   return list;
 }
 
 static SLang_Object_Type *find_nth_element (SLang_List_Type *list, SLindex_Type nth, Chunk_Type **cp)
@@ -246,14 +260,13 @@ static SLang_Object_Type *find_nth_element (SLang_List_Type *list, SLindex_Type
    return c->elements + (nth - next_n);
 }
 
-static int pop_list (SLang_MMT_Type **mmtp, SLang_List_Type **list)
+static int pop_list (SLang_List_Type **list)
 {
-   SLang_MMT_Type *mmt;
-
-   if (NULL == (mmt = SLang_pop_mmt (SLANG_LIST_TYPE)))
-     return -1;
-   *list = (SLang_List_Type *)SLang_object_from_mmt (mmt);
-   *mmtp = mmt;
+   if (-1 == SLclass_pop_ptr_obj (SLANG_LIST_TYPE, (VOID_STAR *) list))
+     {
+	*list = NULL;
+	return -1;
+     }
    return 0;
 }
 
@@ -268,7 +281,7 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, SLindex_Type indx_a
    SLang_Object_Type *obj, *obj_max, *new_obj, *new_obj_max;
 
    if (length == 0)
-     return allocate_list ();
+     return allocate_list (0);
 
    if ((indx_a < 0) || (indx_a + (length - 1) >= list->length))
      {
@@ -276,18 +289,18 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, SLindex_Type indx_a
 	return NULL;
      }
 
-   if (NULL == (new_list = allocate_list ()))
+   if (NULL == (new_list = allocate_list (length)))
      return NULL;
 
-   if (-1 == make_chunk_chain (length, &new_list->first, &new_list->last))
+   if (-1 == make_chunk_chain (length, &new_list->first, &new_list->last, list->default_chunk_size))
      {
-	delete_list (new_list);
+	free_list (new_list);
 	return NULL;
      }
    obj = find_nth_element (list, indx_a, &c);
    if (obj == NULL)
      {
-	delete_list (new_list);
+	free_list (new_list);
 	return NULL;
      }
    obj_max = c->elements + c->num_elements;
@@ -295,7 +308,7 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, SLindex_Type indx_a
    new_list->length = length;
    new_c = new_list->first;
    new_obj = new_c->elements;
-   new_obj_max = new_obj + CHUNK_SIZE;
+   new_obj_max = new_obj + new_c->chunk_size;
 
    for (i = 0; i < length; i++)
      {
@@ -309,13 +322,13 @@ static SLang_List_Type *make_sublist (SLang_List_Type *list, SLindex_Type indx_a
 	  {
 	     new_c = new_c->next;
 	     new_obj = new_c->elements;
-	     new_obj_max = new_obj + CHUNK_SIZE;
+	     new_obj_max = new_obj + new_c->chunk_size;
 	  }
 
 	if ((-1 == _pSLpush_slang_obj (obj))
 	    || (-1 == SLang_pop (new_obj)))
 	  {
-	     delete_list (new_list);
+	     free_list (new_list);
 	     return NULL;
 	  }
 
@@ -387,20 +400,22 @@ static int insert_element (SLang_List_Type *list, SLang_Object_Type *obj, SLinde
 {
    Chunk_Type *c, *c1;
    SLang_Object_Type *elem;
-   int num;
+   SLindex_Type num;
    int equality_ok = 0;
 
    if (indx == 0)
      {
 	c = list->first;
 	if ((c != NULL)
-	    && (c->num_elements < CHUNK_SIZE))
+	    && (c->num_elements < c->chunk_size))
 	  {
 	     slide_right (c, 0);
 	     c->elements[0] = *obj;
 	     goto the_return;
 	  }
-	if (NULL == (c = new_chunk ()))
+	if (list->default_chunk_size < DEFAULT_CHUNK_SIZE)
+	  list->default_chunk_size *= 2;
+	if (NULL == (c = new_chunk (list->default_chunk_size)))
 	  return -1;
 
 	c->next = list->first;
@@ -418,12 +433,16 @@ static int insert_element (SLang_List_Type *list, SLang_Object_Type *obj, SLinde
    if (indx == list->length)
      {
 	c = list->last;
-	if (c->num_elements < CHUNK_SIZE)
+	if (c->num_elements < c->chunk_size)
 	  {
 	     c->elements[c->num_elements] = *obj;
 	     goto the_return;
 	  }
-	if (NULL == (c = new_chunk ()))
+
+	if (list->default_chunk_size < DEFAULT_CHUNK_SIZE)
+	  list->default_chunk_size *= 2;
+
+	if (NULL == (c = new_chunk (list->default_chunk_size)))
 	  return -1;
 	c->prev = list->last;
 	list->last->next = c;
@@ -435,18 +454,21 @@ static int insert_element (SLang_List_Type *list, SLang_Object_Type *obj, SLinde
    if (NULL == (elem = find_nth_element (list, indx, &c)))
      return -1;
 
-   if (c->num_elements < CHUNK_SIZE)
+   if (c->num_elements < c->chunk_size)
      {
 	slide_right (c, elem - c->elements);
 	*elem = *obj;
 	goto the_return;
      }
 
-   if (NULL == (c1 = new_chunk ()))
+   if (list->default_chunk_size < DEFAULT_CHUNK_SIZE)
+     list->default_chunk_size *= 2;
+
+   if (NULL == (c1 = new_chunk (list->default_chunk_size)))
      return -1;
 
-   num = CHUNK_SIZE - (elem - c->elements);
-   if (num == CHUNK_SIZE)
+   num = c->chunk_size - (elem - c->elements);
+   if (num == c->chunk_size)
      {
 	c1->next = c;
 	c1->prev = c->prev;
@@ -492,7 +514,7 @@ the_return:
  * Upon failure, calling routine should free obj.
  */
 
-static int pop_insert_append_args (SLang_MMT_Type **mmtp, SLang_List_Type **listp,
+static int pop_insert_append_args (SLang_List_Type **listp,
 				   SLang_Object_Type *obj, int *indx)
 {
    if (SLang_Num_Function_Args == 3)
@@ -502,7 +524,7 @@ static int pop_insert_append_args (SLang_MMT_Type **mmtp, SLang_List_Type **list
      }
    if (-1 == SLang_pop (obj))
      return -1;
-   if (-1 == pop_list (mmtp, listp))
+   if (-1 == pop_list (listp))
      {
 	SLang_free_object (obj);
 	return -1;
@@ -514,11 +536,10 @@ static void list_insert_elem (void)
 {
    int indx;
    SLang_Object_Type obj;
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list;
 
    indx = 0;
-   if (-1 == pop_insert_append_args (&mmt, &list, &obj, &indx))
+   if (-1 == pop_insert_append_args (&list, &obj, &indx))
      return;
 
    if (indx < 0)
@@ -527,18 +548,17 @@ static void list_insert_elem (void)
    if (-1 == insert_element (list, &obj, indx))
      SLang_free_object (&obj);
 
-   SLang_free_mmt (mmt);
+   free_list (list);
 }
 
 static void list_append_elem (void)
 {
    int indx;
    SLang_Object_Type obj;
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list;
 
    indx = -1;
-   if (-1 == pop_insert_append_args (&mmt, &list, &obj, &indx))
+   if (-1 == pop_insert_append_args (&list, &obj, &indx))
      return;
 
    if (indx < 0)
@@ -547,34 +567,40 @@ static void list_append_elem (void)
    if (-1 == insert_element (list, &obj, indx+1))
      SLang_free_object (&obj);
 
-   SLang_free_mmt (mmt);
+   free_list (list);
 }
 
-static int push_list (SLang_List_Type *list)
+static int push_list (SLang_List_Type *list, int free_flag)
 {
-   SLang_MMT_Type *mmt;
-   if (NULL == (mmt = SLang_create_mmt (SLANG_LIST_TYPE, (VOID_STAR) list)))
+   /* SLclass_push_ptr_obj does not do memory management */
+   if (-1 == SLclass_push_ptr_obj (SLANG_LIST_TYPE, (VOID_STAR)list))
      {
-	delete_list (list);
+	if (free_flag) free_list (list);
 	return -1;
      }
 
-   if (-1 == SLang_push_mmt (mmt))
-     {
-	SLang_free_mmt (mmt);
-	return -1;
-     }
+   if (free_flag == 0)
+     list->ref_count++;
+
    return 0;
 }
 
 static void list_new (void)
 {
    SLang_List_Type *list;
+   int len = DEFAULT_CHUNK_SIZE;
 
-   if (NULL == (list = allocate_list ()))
+   if (SLang_Num_Function_Args == 1)
+     {
+	if (-1 == SLang_pop_integer (&len))
+	  return;
+	if (len <= 0) len = DEFAULT_CHUNK_SIZE;
+     }
+
+   if (NULL == (list = allocate_list (len)))
      return;
 
-   (void) push_list (list);
+   (void) push_list (list, 1);
 }
 
 static void list_reverse (SLang_List_Type *list)
@@ -628,7 +654,6 @@ static int list_pop_nth (SLang_List_Type *list, SLindex_Type indx)
 static void list_pop (void)
 {
    int indx = 0;
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list;
 
    if (SLang_Num_Function_Args == 2)
@@ -636,18 +661,18 @@ static void list_pop (void)
 	if (-1 == SLang_pop_integer (&indx))
 	  return;
      }
-   if (-1 == pop_list (&mmt, &list))
+   if (-1 == pop_list (&list))
      return;
 
    (void) list_pop_nth (list, indx);
-   SLang_free_mmt (mmt);
+   free_list (list);
 }
 
 static int pop_as_list_internal (unsigned int count)
 {
    SLang_List_Type *list;
 
-   if (NULL == (list = allocate_list ()))
+   if (NULL == (list = allocate_list (count)))
      return -1;
 
    while (count)
@@ -665,10 +690,10 @@ static int pop_as_list_internal (unsigned int count)
 
 	count--;
      }
-   return push_list (list);
+   return push_list (list, 1);
 
-   return_error:
-   delete_list (list);
+return_error:
+   free_list (list);
    return -1;
 }
 
@@ -723,14 +748,13 @@ static int l2a_push_callback (VOID_STAR vlist, SLuindex_Type i)
 static void list_to_array (void)
 {
    SLang_List_Type *list;
-   SLang_MMT_Type *mmt;
    SLtype type = 0;
 
    if ((SLang_Num_Function_Args == 2)
        && (-1 == SLang_pop_datatype (&type)))
      return;
 
-   if (-1 == pop_list (&mmt, &list))
+   if (-1 == pop_list (&list))
      return;
 
    (void) _pSLarray_convert_to_array ((VOID_STAR) list,
@@ -738,7 +762,7 @@ static void list_to_array (void)
 				      l2a_push_callback,
 				      list->length, type);
 
-   SLang_free_mmt (mmt);
+   free_list (list);
 }
 
 /* joins  l2 onto l1 */
@@ -789,11 +813,11 @@ static void list_concat (SLang_List_Type *l1, SLang_List_Type *l2)
 
    if (-1 == list_join_internal (list, l2))
      {
-	delete_list (list);
+	free_list (list);
 	return;
      }
 
-  (void) push_list (list);	       /* will delete upon failure */
+  (void) push_list (list, 1);	       /* will delete upon failure */
 }
 
 #define L SLANG_LIST_TYPE
@@ -823,22 +847,22 @@ static int list_length (SLtype type, VOID_STAR v, SLuindex_Type *len)
    SLang_List_Type *list;
 
    (void) type;
-   list = (SLang_List_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)v);
+   list = *(SLang_List_Type **) v;
    *len = list->length;
    return 0;
 }
 
-static int list_dereference (SLtype type, VOID_STAR addr)
+static int cl_list_dereference (SLtype type, VOID_STAR addr)
 {
    SLang_List_Type *list;
    (void) type;
 
-   list = (SLang_List_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)addr);
+   list = *(SLang_List_Type **) addr;
 
    if (NULL == (list = make_sublist (list, 0, list->length)))
      return -1;
 
-   return push_list (list);	       /* will delete upon failure */
+   return push_list (list, 1);	       /* will delete upon failure */
 }
 
 static char *string_method (SLtype type, VOID_STAR p)
@@ -847,7 +871,7 @@ static char *string_method (SLtype type, VOID_STAR p)
    char buf[256];
 
    (void) type;
-   list = (SLang_List_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)p);
+   list = *(SLang_List_Type **) p;
 
    sprintf (buf, "List_Type with %ld elements", (long)list->length);
    return SLmake_string (buf);
@@ -862,8 +886,8 @@ static int eqs_method (SLtype a_type, VOID_STAR pa, SLtype b_type, VOID_STAR pb)
    if ((a_type != SLANG_LIST_TYPE) || (b_type != SLANG_LIST_TYPE))
      return 0;
 
-   a = (SLang_List_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)pa);
-   b = (SLang_List_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)pb);
+   a = *(SLang_List_Type **) pa;
+   b = *(SLang_List_Type **) pb;
 
    if (a == b)
      return 1;
@@ -913,7 +937,6 @@ static int eqs_method (SLtype a_type, VOID_STAR pa, SLtype b_type, VOID_STAR pb)
 struct _pSLang_Foreach_Context_Type
 {
    SLang_List_Type *list;
-   SLang_MMT_Type *mmt;
    int next_index;
 };
 
@@ -935,7 +958,7 @@ cl_foreach_open (SLtype type, unsigned int num)
    if (NULL == (c = (SLang_Foreach_Context_Type *) SLcalloc (1, sizeof (SLang_Foreach_Context_Type))))
      return NULL;
 
-   if (-1 == pop_list (&c->mmt, &c->list))
+   if (-1 == pop_list (&c->list))
      {
 	SLfree ((char *) c);
 	return NULL;
@@ -948,7 +971,7 @@ static void cl_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    if (c == NULL) return;
-   SLang_free_mmt (c->mmt);
+   free_list (c->list);
    SLfree ((char *) c);
 }
 
@@ -972,20 +995,21 @@ static int cl_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 }
 
 static int pop_list_and_index (unsigned int num_indices,
-			       SLang_MMT_Type **mmtp, SLang_List_Type **listp,
+			       SLang_List_Type **listp,
 			       SLang_Array_Type **ind_atp,
 			       SLindex_Type *indx)
 {
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list;
 
-   if (-1 == pop_list (&mmt, &list))
+   *listp = NULL;
+
+   if (-1 == pop_list (&list))
      return -1;
 
    if (num_indices != 1)
      {
 	_pSLang_verror (SL_InvalidParm_Error, "List_Type objects are limited to a single index");
-	SLang_free_mmt (mmt);
+	free_list (list);
 	return -1;
      }
 
@@ -994,7 +1018,7 @@ static int pop_list_and_index (unsigned int num_indices,
      {
 	if (-1 == SLang_pop_array_index (indx))
 	  {
-	     SLang_free_mmt (mmt);
+	     free_list (list);
 	     return -1;
 	  }
      }
@@ -1002,30 +1026,28 @@ static int pop_list_and_index (unsigned int num_indices,
      {
 	if (-1 == _pSLarray_pop_index (list->length, ind_atp, indx))
 	  {
-	     SLang_free_mmt (mmt);
+	     free_list (list);
 	     return -1;
 	  }
      }
 
    *listp = list;
-   *mmtp = mmt;
    return 0;
 }
 
 /* FIXME: Extend this to allow an index array */
 static int _pSLlist_aget (SLtype type, unsigned int num_indices)
 {
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list, *new_list;
    SLang_Object_Type *obj;
    SLang_Array_Type *ind_at;
    SLindex_Type indx, *idx_data;
-   SLuindex_Type i, num, list_len;
+   SLuindex_Type i, num;
    int ret;
 
    (void) type;
 
-   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &ind_at, &indx))
+   if (-1 == pop_list_and_index (num_indices, &list, &ind_at, &indx))
      return -1;
 
    ret = -1;
@@ -1035,15 +1057,15 @@ static int _pSLlist_aget (SLtype type, unsigned int num_indices)
 	if (obj != NULL)
 	  ret = _pSLpush_slang_obj (obj);
 
-	SLang_free_mmt (mmt);
+	free_list (list);
 	return ret;
      }
 
-   if (NULL == (new_list = allocate_list ()))
+   num = ind_at->num_elements;
+
+   if (NULL == (new_list = allocate_list (num)))
      goto free_and_return;
 
-   list_len = list->length;
-   num = ind_at->num_elements;
    idx_data = (SLindex_Type *)ind_at->data;
    for (i = 0; i < num; i++)
      {
@@ -1064,14 +1086,14 @@ static int _pSLlist_aget (SLtype type, unsigned int num_indices)
 	  }
      }
 
-   ret = push_list (new_list);	       /* frees upon error */
+   ret = push_list (new_list, 1);	       /* frees upon error */
    new_list = NULL;
 
 free_and_return:
 
    if (new_list != NULL)
-     delete_list (new_list);
-   SLang_free_mmt (mmt);
+     free_list (new_list);
+   free_list (list);
    SLang_free_array (ind_at);
    return ret;
 }
@@ -1090,44 +1112,42 @@ static int aput_object (SLang_List_Type *list, SLindex_Type indx, SLang_Object_T
 
 static int _pSLlist_aput (SLtype type, unsigned int num_indices)
 {
-   SLang_MMT_Type *mmt;
    SLang_List_Type *list;
    SLang_Object_Type obj;
    SLang_Array_Type *ind_at;
    SLindex_Type indx, *idx_data;
-   SLuindex_Type i, num, list_len;
+   SLuindex_Type i, num;
    int ret;
 
    (void) type;
 
-   if (-1 == pop_list_and_index (num_indices, &mmt, &list, &ind_at, &indx))
+   if (-1 == pop_list_and_index (num_indices, &list, &ind_at, &indx))
      return -1;
 
    if (ind_at == NULL)
      {
 	if (-1 == SLang_pop (&obj))
 	  {
-	     SLang_free_mmt (mmt);
+	     free_list (list);
 	     return -1;
 	  }
 
 	if (-1 == aput_object (list, indx, &obj))
 	  {
 	     SLang_free_object (&obj);
-	     SLang_free_mmt (mmt);
+	     free_list (list);
 	     return -1;
 	  }
-	SLang_free_mmt (mmt);
+	free_list (list);
 	return 0;
      }
 
    idx_data = (SLindex_Type *)ind_at->data;
    num = ind_at->num_elements;
-   list_len = list->length;
 
    if (-1 == SLang_pop (&obj))
      {
-	SLang_free_mmt (mmt);
+	free_list (list);
 	SLang_free_array (ind_at);
 	return -1;
      }
@@ -1164,11 +1184,9 @@ static int _pSLlist_aput (SLtype type, unsigned int num_indices)
 
    if (obj.o_data_type == SLANG_LIST_TYPE)
      {
-	SLang_MMT_Type *mmt2;
 	SLang_List_Type *list2;
 
-	mmt2 = obj.v.ref;
-	if (NULL == (list2 = (SLang_List_Type *)SLang_object_from_mmt (mmt2)))
+	if (NULL == (list2 = (SLang_List_Type *)obj.v.ptr_val))
 	  goto free_and_return;
 
 	if (list2->length != (SLindex_Type)num)
@@ -1220,10 +1238,23 @@ free_and_return:
 
    SLang_free_object (&obj);
    SLang_free_array (ind_at);
-   SLang_free_mmt (mmt);
+   free_list (list);
    return ret;
 }
 
+
+static void cl_list_destroy (SLtype type, VOID_STAR ptr)
+{
+   (void) type;
+   free_list (*(SLang_List_Type **) ptr);
+}
+
+static int cl_list_push (SLtype type, VOID_STAR ptr)
+{
+   (void) type;
+   return push_list (*(SLang_List_Type **)ptr, 0);
+}
+
 int _pSLang_init_sllist (void)
 {
    SLang_Class_Type *cl;
@@ -1234,10 +1265,11 @@ int _pSLang_init_sllist (void)
    if (NULL == (cl = SLclass_allocate_class ("List_Type")))
      return -1;
 
-   (void) SLclass_set_destroy_function (cl, list_destroy);
+   (void) SLclass_set_destroy_function (cl, cl_list_destroy);
+   (void) SLclass_set_push_function (cl, cl_list_push);
    (void) SLclass_set_aput_function (cl, _pSLlist_aput);
    (void) SLclass_set_aget_function (cl, _pSLlist_aget);
-   (void) SLclass_set_deref_function (cl, list_dereference);
+   (void) SLclass_set_deref_function (cl, cl_list_dereference);
    (void) SLclass_set_string_function (cl, string_method);
    (void) SLclass_set_eqs_function (cl, eqs_method);
    (void) SLclass_set_is_container (cl, 1);
@@ -1248,7 +1280,7 @@ int _pSLang_init_sllist (void)
    cl->cl_foreach_close = cl_foreach_close;
    cl->cl_foreach = cl_foreach;
 
-   if (-1 == SLclass_register_class (cl, SLANG_LIST_TYPE, sizeof (SLang_List_Type), SLANG_CLASS_TYPE_MMT))
+   if (-1 == SLclass_register_class (cl, SLANG_LIST_TYPE, sizeof (SLang_List_Type), SLANG_CLASS_TYPE_PTR))
      return -1;
 
    if (-1 == SLadd_intrin_fun_table (Intrin_Table, NULL))
@@ -1257,3 +1289,61 @@ int _pSLang_init_sllist (void)
    return 0;
 }
 
+SLang_List_Type *SLang_create_list (SLindex_Type chunk_size)
+{
+   return allocate_list (chunk_size);
+}
+
+int SLang_list_append (SLang_List_Type *list, int indx)
+{
+   SLang_Object_Type obj;
+
+   if (-1 == SLang_pop (&obj))
+     return -1;
+
+   if (indx < 0)
+     indx += list->length;
+
+   if (-1 == insert_element (list, &obj, indx+1))
+     {
+	SLang_free_object (&obj);
+	return -1;
+     }
+   return 0;
+}
+
+int SLang_list_insert (SLang_List_Type *list, int indx)
+{
+   SLang_Object_Type obj;
+
+   if (-1 == SLang_pop (&obj))
+     return -1;
+
+   if (indx < 0)
+     indx += list->length;
+
+   if (-1 == insert_element (list, &obj, indx))
+     {
+	SLang_free_object (&obj);
+	return -1;
+     }
+   return 0;
+}
+
+int SLang_pop_list (SLang_List_Type **list)
+{
+   return pop_list (list);
+}
+
+int SLang_push_list (SLang_List_Type *list, int free_flag)
+{
+   if (list == NULL)
+     return SLang_push_null ();
+
+   return push_list (list, free_flag);
+}
+
+void SLang_free_list (SLang_List_Type *list)
+{
+   free_list (list);
+}
