@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2010 John E. Davis
+Copyright (C) 2004-2016 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -28,8 +28,8 @@ typedef struct _pSLstring_Type
 {
    struct _pSLstring_Type *next;
    unsigned int ref_count;
-   unsigned long hash;
-   unsigned int len;
+   SLstr_Hash_Type hash;
+   size_t len;
    char bytes [1];
 }
 SLstring_Type;
@@ -54,7 +54,7 @@ static SLCONST char *Deleted_String = "*deleted*";
 static Cached_String_Type Cached_Strings [NUM_CACHED_STRINGS];
 
 #define GET_CACHED_STRING(s) \
-   (Cached_Strings + (unsigned int)(((unsigned long) (s)) % NUM_CACHED_STRINGS))
+   (Cached_Strings + (unsigned int)(((size_t) (s)) % NUM_CACHED_STRINGS))
 
 _INLINE_
 static void cache_string (SLstring_Type *sls)
@@ -67,7 +67,7 @@ static void cache_string (SLstring_Type *sls)
 }
 
 _INLINE_
-static void uncache_string (char *s)
+static void uncache_string (SLCONST char *s)
 {
    Cached_String_Type *cs;
 
@@ -102,7 +102,7 @@ static void uncache_string (char *s)
 }
 
 _INLINE_
-unsigned long _pSLstring_hash (SLCONST unsigned char *s, SLCONST unsigned char *smax)
+SLstr_Hash_Type _pSLstring_hash (SLCONST unsigned char *s, SLCONST unsigned char *smax)
 {
    register _pSLuint32_Type a,b,c;
    unsigned int length = (unsigned int)(smax - s);
@@ -142,7 +142,7 @@ unsigned long _pSLstring_hash (SLCONST unsigned char *s, SLCONST unsigned char *
    mix(a,b,c);
 
    /*-------------------------------------------- report the result */
-   return (unsigned long) c;
+   return (SLstr_Hash_Type) c;
 }
 #else
 _INLINE_
@@ -177,7 +177,7 @@ unsigned long _pSLstring_hash (SLCONST unsigned char *s, SLCONST unsigned char *
    return h;
 }
 #endif
-unsigned long _pSLcompute_string_hash (SLCONST char *s)
+SLstr_Hash_Type SLcompute_string_hash (SLCONST char *s)
 {
 #if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
@@ -190,10 +190,10 @@ unsigned long _pSLcompute_string_hash (SLCONST char *s)
 }
 
 _INLINE_
-static SLstring_Type *find_slstring (SLCONST char *s, unsigned long hash)
+static SLstring_Type *find_slstring (SLCONST char *s, SLstr_Hash_Type hash)
 {
    SLstring_Type *sls, *prev;
-   unsigned int idx = MAP_HASH_TO_INDEX(hash);
+   size_t idx = MAP_HASH_TO_INDEX(hash);
 
    sls = String_Hash_Table [idx];
    if ((sls == NULL) || (sls->bytes == s)) return sls;
@@ -225,7 +225,7 @@ static SLstring_Type *find_slstring (SLCONST char *s, unsigned long hash)
 
 _INLINE_
 /* This routine works with any (long) string */
-static SLstring_Type *find_string (SLCONST char *s, unsigned int len, unsigned long hash)
+static SLstring_Type *find_string (SLCONST char *s, unsigned int len, SLstr_Hash_Type hash)
 {
    SLstring_Type *sls;
 
@@ -289,7 +289,7 @@ _INLINE_
 static void free_sls (SLstring_Type *sls)
 {
 #if SLANG_OPTIMIZE_FOR_SPEED
-   unsigned int len = sls->len;
+   size_t len = sls->len;
    if ((len < MAX_FREE_STORE_LEN)
        && (SLS_Free_Store[len] == NULL))
      {
@@ -301,7 +301,7 @@ static void free_sls (SLstring_Type *sls)
 }
 
 _INLINE_
-static char *create_long_string (SLCONST char *s, unsigned int len, unsigned long hash)
+static char *create_long_string (SLCONST char *s, size_t len, SLstr_Hash_Type hash)
 {
    SLstring_Type *sls;
 
@@ -353,9 +353,9 @@ static char *create_short_string (SLCONST char *s, unsigned int len)
 
 /* s cannot be NULL */
 _INLINE_
-static SLstr_Type *create_nstring (SLCONST char *s, unsigned int len, unsigned long *hash_ptr)
+static SLstr_Type *create_nstring (SLCONST char *s, size_t len, SLstr_Hash_Type *hash_ptr)
 {
-   unsigned long hash;
+   SLstr_Hash_Type hash;
 
    if (len < 2)
      return create_short_string (s, len);
@@ -368,15 +368,15 @@ static SLstr_Type *create_nstring (SLCONST char *s, unsigned int len, unsigned l
 
 SLstr_Type *SLang_create_nslstring (SLFUTURE_CONST char *s, SLstrlen_Type len)
 {
-   unsigned long hash;
+   SLstr_Hash_Type hash;
    if (s == NULL)
      return NULL;
    return create_nstring (s, len, &hash);
 }
 
-char *_pSLstring_make_hashed_string (SLCONST char *s, unsigned int len, unsigned long *hashptr)
+char *_pSLstring_make_hashed_string (SLCONST char *s, SLstrlen_Type len, SLstr_Hash_Type *hashptr)
 {
-   unsigned long hash;
+   SLstr_Hash_Type hash;
 
    if (s == NULL) return NULL;
 
@@ -389,9 +389,9 @@ char *_pSLstring_make_hashed_string (SLCONST char *s, unsigned int len, unsigned
    return create_long_string (s, len, hash);
 }
 
-char *_pSLstring_dup_hashed_string (SLCONST char *s, unsigned long hash)
+char *_pSLstring_dup_hashed_string (SLCONST char *s, SLstr_Hash_Type hash)
 {
-   unsigned int len;
+   size_t len;
 #if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
 
@@ -451,7 +451,7 @@ SLCONST char *_pSLstring_dup_slstring (SLCONST char *s)
 static void free_sls_string (SLstring_Type *sls)
 {
    SLstring_Type *sls1, *prev;
-   unsigned long hash = sls->hash;
+   SLstr_Hash_Type hash = sls->hash;
 
    hash = MAP_HASH_TO_INDEX(hash);
 
@@ -475,7 +475,7 @@ static void free_sls_string (SLstring_Type *sls)
 }
 
 _INLINE_
-static void free_long_string (char *s, unsigned long hash)
+static void free_long_string (SLCONST char *s, SLstr_Hash_Type hash)
 {
    SLstring_Type *sls;
 
@@ -500,10 +500,10 @@ static void free_long_string (char *s, unsigned long hash)
 }
 
 /* This routine may be passed NULL-- it is not an error. */
-void SLang_free_slstring (char *s)
+void SLang_free_slstring (SLCONST char *s)
 {
-   unsigned long hash;
-   unsigned int len;
+   SLstr_Hash_Type hash;
+   size_t len;
 #if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
 #endif
@@ -538,7 +538,7 @@ void SLang_free_slstring (char *s)
 
 char *SLang_create_slstring (SLFUTURE_CONST char *s)
 {
-   unsigned long hash;
+   SLstr_Hash_Type hash;
 #if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
 #endif
@@ -556,13 +556,13 @@ char *SLang_create_slstring (SLFUTURE_CONST char *s)
    return create_nstring (s, strlen (s), &hash);
 }
 
-void _pSLfree_hashed_string (char *s, unsigned int len, unsigned long hash)
+void _pSLfree_hashed_string (SLCONST char *s, size_t len, SLstr_Hash_Type hash)
 {
    if ((s == NULL) || (len < 2)) return;
    free_long_string (s, hash);
 }
 
-char *_pSLallocate_slstring (unsigned int len)
+char *_pSLallocate_slstring (size_t len)
 {
    SLstring_Type *sls = allocate_sls (len);
    if (sls == NULL)
@@ -572,7 +572,7 @@ char *_pSLallocate_slstring (unsigned int len)
    return sls->bytes;
 }
 
-void _pSLunallocate_slstring (char *s, unsigned int len)
+void _pSLunallocate_slstring (char *s, size_t len)
 {
    SLstring_Type *sls;
 
@@ -585,9 +585,9 @@ void _pSLunallocate_slstring (char *s, unsigned int len)
 }
 
 /* frees s upon error */
-char *_pSLcreate_via_alloced_slstring (char *s, unsigned int len)
+char *_pSLcreate_via_alloced_slstring (char *s, size_t len)
 {
-   unsigned long hash;
+   SLstr_Hash_Type hash;
    SLstring_Type *sls;
 
    if (s == NULL)
@@ -654,7 +654,7 @@ char *SLang_concat_slstrings (char *a, char *b)
 }
 
 /* This routine is assumed to work even if s is not an slstring */
-unsigned int _pSLstring_bytelen (SLstr_Type *s)
+size_t _pSLstring_bytelen (SLCONST SLstr_Type *s)
 {
 #if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
@@ -708,7 +708,7 @@ void _pSLang_free_slstring (SLstr_Type *s)
 }
 
 /* An SLstring is required */
-unsigned long _pSLstring_get_hash (SLstr_Type *s)
+SLstr_Hash_Type _pSLstring_get_hash (SLstr_Type *s)
 {
    SLstring_Type *sls;
 
