@@ -1,7 +1,7 @@
 /* -*- mode: C; mode: fold; -*- */
 /* string manipulation functions for S-Lang. */
 /*
-Copyright (C) 2004-2010 John E. Davis
+Copyright (C) 2004-2016 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -160,7 +160,7 @@ static void strcat_cmd (void) /*{{{*/
 
    if (nargs <= 10)
      ptrs = ptrs_buf;
-   else if (NULL == (ptrs = (char **)SLmalloc (nargs * sizeof (char *))))
+   else if (NULL == (ptrs = (char **)_SLcalloc (nargs, sizeof (char *))))
      return;
 
    memset ((char *) ptrs, 0, sizeof (char *) * nargs);
@@ -219,7 +219,7 @@ static int str_replace_cmd_1 (char *orig, char *match, char *rep, unsigned int m
 			      char **new_strp) /*{{{*/
 {
    char *s, *t, *new_str;
-   unsigned int rep_len, match_len, new_len;
+   size_t rep_len, match_len, new_len;
    unsigned int num_replaces;
 
    *new_strp = NULL;
@@ -248,7 +248,7 @@ static int str_replace_cmd_1 (char *orig, char *match, char *rep, unsigned int m
    rep_len = strlen (rep);
 
    new_len = (strlen (orig) - num_replaces * match_len) + num_replaces * rep_len;
-   new_str = SLmalloc (new_len + 1);
+   new_str = (char *)SLmalloc (new_len + 1);
    if (new_str == NULL)
      return -1;
 
@@ -432,7 +432,7 @@ static void str_uncomment_string_cmd (char *str, char *b, char *e) /*{{{*/
 {
    SLuchar_Type *s, *smax, *bmax, *emax;
    SLwchar_Lut_Type *lut;
-   unsigned int len, elen, blen;
+   size_t len, elen, blen;
    SLuchar_Type *etable;
    SLuchar_Type *b1, *e1;
    int ignore_combining = 0;
@@ -448,7 +448,7 @@ static void str_uncomment_string_cmd (char *str, char *b, char *e) /*{{{*/
 	return;
      }
 
-   if (NULL == (etable = (SLuchar_Type *) SLmalloc (blen * (SLUTF8_MAX_MBLEN+1))))
+   if (NULL == (etable = (SLuchar_Type *) _SLcalloc (blen, (SLUTF8_MAX_MBLEN+1))))
      return;
 
    b1 = (SLuchar_Type *) b;
@@ -546,7 +546,7 @@ static void str_quote_string_cmd (char *str, char *quotes, SLwchar_Type *slash_p
 {
    char *q, *q1;
    int slash;
-   unsigned int len;
+   size_t len;
    SLwchar_Lut_Type *lut;
    SLuchar_Type slash_utf8 [SLUTF8_MAX_MBLEN+1];
    SLuchar_Type *s, *smax;
@@ -587,7 +587,7 @@ static void str_quote_string_cmd (char *str, char *quotes, SLwchar_Type *slash_p
      }
    len++;			       /* null terminate */
 
-   if (NULL == (q = SLmalloc(len)))
+   if (NULL == (q = (char *)SLmalloc(len)))
      {
 	SLwchar_free_lut (lut);
 	return;
@@ -626,10 +626,10 @@ static void str_quote_string_cmd (char *str, char *quotes, SLwchar_Type *slash_p
 static void subbytes_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
 {
    int m;
-   unsigned int n;
-   unsigned int lena;
+   size_t n;
+   size_t lena;
 
-   n = (unsigned int) (*n_ptr - 1);
+   n = (*n_ptr - 1);
    m = *m_ptr;
 
    lena = _pSLstring_bytelen (a);
@@ -692,13 +692,13 @@ static void substr_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
 static void strbytesub_cmd (int *nptr, char *chp)
 {
    char *a;
-   unsigned int n;
-   unsigned int lena;
+   size_t n;
+   size_t lena;
 
    if (-1 == SLpop_string (&a))
      return;
 
-   n = (unsigned int) (*nptr-1);
+   n = (*nptr-1);
    lena = strlen (a);
 
    if (n >= lena)
@@ -717,9 +717,9 @@ static void strbytesub_cmd (int *nptr, char *chp)
 static void strsub_cmd (int *nptr, SLwchar_Type *mptr) /*{{{*/
 {
    char *a;
-   unsigned int n;
+   size_t n;
    SLwchar_Type m;
-   unsigned int lena;
+   size_t lena;
    int ignore_combining = 0;
 
    if (-1 == SLpop_string (&a))
@@ -730,7 +730,7 @@ static void strsub_cmd (int *nptr, SLwchar_Type *mptr) /*{{{*/
 
    lena = STRLEN (a, ignore_combining);
 
-   if ((n == 0) || (lena < (unsigned int) n))
+   if ((n == 0) || (lena < n))
      {
 	SLang_set_error (SL_INVALID_PARM);
 	SLfree(a);
@@ -782,7 +782,7 @@ static int pop_wchar (SLwchar_Type *wchp)
 	return 0;
      }
 
-   return SLang_pop_wchar (wchp);
+   return _pSLang_pop_wchar (wchp);
 }
 
 #define ISXXX_INTRIN(name,isxxx) \
@@ -900,7 +900,7 @@ static void strbskipchar_intrin (void)
    if (_pSLinterp_UTF8_Mode == 0)
      {
 	(void) SLang_push_strlen_type (pos-1);
-	(void) SLang_push_uchar (*str0);
+	(void) SLang_push_uchar (*(str0-1));
 	goto free_and_return;
      }
    str1 = SLutf8_bskip_chars (str, str0, 1, NULL, skip_combining);
@@ -1381,6 +1381,9 @@ static int strtrim_internal (int do_beg, int do_end)
      }
    else cd.lut = make_whitespace_lut ();
 
+   if (cd.lut == NULL)
+     return -1;
+
    status = arraymap_str_func_str (func_strtrim, &cd);
    if (free_lut) SLwchar_free_lut (cd.lut);
    return status;
@@ -1403,7 +1406,7 @@ static void strtrim_end_vintrin (void)
 
 static char *func_strup (char *str, void *cd)
 {
-   unsigned int i, len;
+   size_t i, len;
    unsigned char *a;
 
    (void) cd;
@@ -1433,7 +1436,7 @@ static void strup_vintrin (void)
 
 static char *func_strlow (char *str, void *cd)
 {
-   unsigned int i, len;
+   size_t i, len;
    unsigned char *a;
 
    (void) cd;
@@ -1664,7 +1667,7 @@ static char *func_str_delete_chars (char *str, void *cd)
    while (t != tmax)
      {
 	SLuchar_Type *t1;
-	unsigned int len;
+	size_t len;
 
 	t1 = SLwchar_skip_range (lut, t, tmax, ignore_combining, invert);
 	if (t1 == NULL)
@@ -1768,7 +1771,7 @@ static char *func_strcompress (char *str, void *cd) /*{{{*/
    char *c;
    Strcompress_CD_Type *info;
    SLuchar_Type *s, *beg, *end;
-   unsigned int len, pref_len;
+   size_t len, pref_len;
    SLwchar_Lut_Type *lut;
    int ignore_combining = 0;
 
@@ -1852,7 +1855,8 @@ static void strcompress_vintrin (char *white) /*{{{*/
    memcpy ((char *)cd.pref_char_buf, white, cd.pref_len);
    cd.pref_char_buf[cd.pref_len] = 0;
 
-   if (NULL == (cd.lut = SLwchar_strtolut ((SLuchar_Type *)white, 1, 0)))
+   /* No ranges and no character classes in white */
+   if (NULL == (cd.lut = SLwchar_strtolut ((SLuchar_Type *)white, 0, 0)))
      return;
 
    (void) arraymap_str_func_str (&func_strcompress, (void *)&cd);
@@ -1862,9 +1866,12 @@ static void strcompress_vintrin (char *white) /*{{{*/
 
 /*}}}*/
 
+#if defined(__GNUC__)
+# pragma GCC diagnostic ignored "-Wformat-nonliteral"
+#endif
 static char *SLdo_sprintf (char *fmt) /*{{{*/
 {
-   register char *p = fmt, ch;
+   register char *p = fmt;
    char *out = NULL, *outp = NULL;
    char dfmt[1024];	       /* used to hold part of format */
    char *f, *f1;
@@ -1875,7 +1882,7 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 #ifdef HAVE_LONG_LONG
    long long llong_var;
 #endif
-   unsigned int len = 0, malloc_len = 0, dlen;
+   size_t len = 0, malloc_len = 0, dlen;
    int do_free;
    unsigned int guess_size;
 #if SLANG_HAS_FLOAT
@@ -1884,9 +1891,12 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 #endif
    unsigned char uch;
    int use_long = 0, use_alt_format = 0;
+   SLuchar_Type utf8_buf[SLUTF8_MAX_MBLEN+1];
 
    while (1)
      {
+	char ch;
+
 	while ((ch = *p) != 0)
 	  {
 	     if (ch == '%')
@@ -1900,9 +1910,9 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 
 	if (len + dlen >= malloc_len)
 	  {
-	     malloc_len = len + dlen;
-	     if (out == NULL) outp = SLmalloc(malloc_len + 1);
-	     else outp = SLrealloc(out, malloc_len + 1);
+	     malloc_len = len + dlen + 512;
+	     if (out == NULL) outp = (char *)SLmalloc(malloc_len + 1);
+	     else outp = (char *)SLrealloc(out, malloc_len + 1);
 	     if (NULL == outp)
 	       return out;
 	     out = outp;
@@ -1969,7 +1979,7 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 
 	if (want_width)
 	  {
-	     sprintf(f, "%d", width);
+	     sprintf(f, "%u", width);
 	     f += strlen (f);
 	  }
 	precis = 0;
@@ -1993,7 +2003,7 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	       }
 	     if (want_width)
 	       {
-		  sprintf(f, "%d", precis);
+		  sprintf(f, "%u", precis);
 		  f += strlen (f);
 	       }
 	     else precis = 0;
@@ -2071,9 +2081,8 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 #endif
 	       {
 		  SLwchar_Type wc;
-		  SLuchar_Type utf8_buf[SLUTF8_MAX_MBLEN+1];
 
-		  if (-1 == SLang_pop_wchar (&wc))
+		  if (-1 == _pSLang_pop_wchar (&wc))
 		    return out;
 		  if ((_pSLinterp_UTF8_Mode == 0) && (wc <= 0xFF))
 		    {
@@ -2103,21 +2112,25 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	   case 'u':
 	   case 'X':
 	   case 'x':
-	     if (use_long)
-	       {
 #ifdef HAVE_LONG_LONG
-		  if (use_long > 1)
-		    {
-		       if (-1 == SLang_pop_long_long (&llong_var))
-			 return out;
-		       *f++ = 'l';
-		    }
-		  else
-#endif
+	     if (use_long > 1)
+	       {
+		  if (-1 == SLang_pop_long_long (&llong_var))
+		    return out;
+# ifdef __WIN32__
+		  *f++ = 'I'; *f++ = '6'; *f++ = '4';
+# else
+		  *f++ = 'l'; *f++ = 'l';
+# endif
+	       }
+	     else
+#endif				       /* HAVE_LONG_LONG */
+	       if (use_long)
+		 {
 		    if (-1 == SLang_pop_long (&long_var))
 		      return out;
-		  *f++ = 'l';
-	       }
+		    *f++ = 'l';
+		 }
 	     else if (-1 == SLang_pop_int (&int_var))
 	       return out;
 	     break;
@@ -2167,6 +2180,7 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 
 	if (len + guess_size > malloc_len)
 	  {
+	     guess_size += 512;
 	     outp = (char *) SLrealloc(out, len + guess_size + 1);
 	     if (outp == NULL)
 	       {
@@ -2204,12 +2218,15 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 
    if (out != NULL)
      {
-	outp = SLrealloc (out, (unsigned int) (outp - out) + 1);
+	outp = (char *)SLrealloc (out, (unsigned int) (outp - out) + 1);
 	if (outp != NULL) out = outp;
      }
 
    return (out);
 }
+#if defined(__GNUC__)
+# pragma GCC diagnostic warning "-Wformat-nonliteral"
+#endif
 
 /*}}}*/
 
@@ -2360,7 +2377,7 @@ static int is_list_element_cmd (char *list, char *elem, SLwchar_Type *delim_ptr)
 {
    SLuchar_Type delim_utf8[SLUTF8_MAX_MBLEN+1];
    unsigned int delim_len;
-   unsigned int elem_len;
+   size_t elem_len;
    int n;
 
    if (NULL == _pSLinterp_encode_wchar (*delim_ptr, delim_utf8, &delim_len))
@@ -2374,7 +2391,7 @@ static int is_list_element_cmd (char *list, char *elem, SLwchar_Type *delim_ptr)
 
    while (1)
      {
-	unsigned int len;
+	size_t len;
 	char *list_end = strstr (list, (char *)delim_utf8);
 
 	if (list_end == NULL)
@@ -2402,8 +2419,8 @@ static unsigned int Regexp_Match_Byte_Offset;
 static int string_match_internal (char *str, char *pat, int n) /*{{{*/
 {
    char *match;
-   unsigned int len;
-   unsigned int byte_offset;
+   size_t len;
+   size_t byte_offset;
 
    if (Regexp != NULL)
      {
@@ -2466,7 +2483,7 @@ static int string_match_cmd (void)
 
 static int string_match_nth_cmd (int *nptr) /*{{{*/
 {
-   unsigned int ofs, len;
+   SLuindex_Type ofs, len;
 
    if (Regexp == NULL)
      {
@@ -2492,9 +2509,9 @@ static int string_match_nth_cmd (int *nptr) /*{{{*/
 static int string_matches_internal (char *str, char *pat, int n)
 {
    int status;
-   unsigned int i;
-   unsigned int lens[10];
-   unsigned int offsets[10];
+   SLuindex_Type i;
+   SLstrlen_Type lens[10];
+   SLstrlen_Type offsets[10];
    char **strs;
    SLindex_Type num;
    SLang_Array_Type *at;
@@ -2545,12 +2562,12 @@ static void string_matches_cmd (void)
 }
 
 /* UTF-8 ok */
-static char *create_delimited_string (char **list, unsigned int n,
+static char *create_delimited_string (char **list, size_t n,
 				      char *delim)
 {
-   unsigned int len, dlen;
-   unsigned int i;
-   unsigned int num;
+   size_t len, dlen;
+   size_t i;
+   size_t num;
    char *str, *s;
 
    len = 1;			       /* allow room for \0 char */
@@ -2566,7 +2583,7 @@ static char *create_delimited_string (char **list, unsigned int n,
    if (num > 1)
      len += (num - 1) * dlen;
 
-   if (NULL == (str = SLmalloc (len)))
+   if (NULL == (str = (char *)SLmalloc (len)))
      return NULL;
 
    *str = 0;
@@ -2575,7 +2592,7 @@ static char *create_delimited_string (char **list, unsigned int n,
 
    while (num > 1)
      {
-	unsigned int len2;
+	size_t len2;
 
 	while (list[i] == NULL)
 	  i++;
@@ -2611,7 +2628,7 @@ static void create_delimited_string_cmd (int *nptr)
 
    n = 1 + (unsigned int) *nptr;       /* n includes delimiter */
 
-   if (NULL == (strings = (char **)SLmalloc (n * sizeof (char *))))
+   if (NULL == (strings = (char **)_SLcalloc (n, sizeof (char *))))
      {
 	SLdo_pop_n (n);
 	return;
@@ -2646,7 +2663,7 @@ static void strjoin_cmd (void)
    if (SLang_Num_Function_Args == 1)
      {
 	free_delim = 0;
-	delim = "";
+	delim = (char *)"";
      }
    else
      {
@@ -2666,7 +2683,7 @@ static void strjoin_cmd (void)
    (void) SLang_push_malloced_string (str);   /* NULL Ok */
 }
 
-static unsigned int count_char_occurances (char *str, SLwchar_Type *wchp)
+static unsigned int count_char_occurrences (char *str, SLwchar_Type *wchp)
 {
    SLwchar_Type wch = *wchp;
    SLuchar_Type wch_utf8[SLUTF8_MAX_MBLEN+1];
@@ -2721,7 +2738,7 @@ static void glob_to_regexp (char *glob)
    char ch;
 
    len = _pSLstring_bytelen (glob);
-   pat = SLmalloc (2*len + 8);
+   pat = (char *)SLmalloc (2*len + 8);
    if (pat == NULL)
      return;
 
@@ -2906,6 +2923,137 @@ free_and_return:
    SLang_free_slstring (chars);
 }
 
+static int copy_strlen_type_to_index_type (SLstrlen_Type a, SLindex_Type *bp)
+{
+   *bp = (SLindex_Type) a;
+   if ((SLstrlen_Type)*bp != a)
+     {
+	SLang_verror (SL_TypeMismatch_Error, "%s", "SLstrlen_Type value to large for conversion to SLindex_Type");
+	return -1;
+     }
+   return 0;
+}
+
+static void string_to_wchars (unsigned char *str)
+{
+   SLstrlen_Type i, len;
+   SLang_Array_Type *at;
+   _pSLint32_Type *data;
+   unsigned char *strmax;
+   SLindex_Type dims[1];
+
+   len = SLutf8_strlen (str, 0);
+   if (-1 == copy_strlen_type_to_index_type (len, dims))
+     return;
+
+   if (NULL == (at = SLang_create_array (_pSLANG_INT32_TYPE, 0, NULL, dims, 1)))
+     return;
+
+   i = 0;
+   strmax = str + _pSLstring_bytelen ((const char *)str);
+   data = (_pSLint32_Type *)at->data;
+   while (str < strmax)
+     {
+	unsigned char *s;
+	SLwchar_Type wch;
+
+	if (*str < 0x80)
+	  {
+	     data[i++] = *str++;
+	     continue;
+	  }
+
+	s = SLutf8_decode (str, strmax, &wch, NULL);
+	if (s == NULL)
+	  {
+	     data[i++] = -(int)(*str);
+	     str++;
+	     continue;
+	  }
+	data[i++] = (_pSLint32_Type)wch;
+	str = s;
+     }
+
+   (void) SLang_push_array (at, 1);
+}
+
+static void wchars_to_string (void)
+{
+   SLindex_Type i, n;
+   SLang_Array_Type *at;
+   _pSLint32_Type *data;
+   unsigned char *buf, *b, *bmax;
+   SLstrlen_Type buflen;
+
+   if (-1 == SLang_pop_array_of_type (&at, _pSLANG_INT32_TYPE))
+     return;
+
+   buflen = n = at->num_elements;
+   buf = (unsigned char *)SLmalloc(buflen+1);
+   if (buf == NULL)
+     {
+	SLang_free_array (at);
+	return;
+     }
+
+   data = (_pSLint32_Type *) at->data;
+   b = buf;
+   bmax = b + buflen;
+
+   i = 0;
+   for (i = 0; i < n; i++)
+     {
+	SLstrlen_Type dlen;
+	_pSLint32_Type wch;
+
+	wch = data[i];
+	if ((wch < 0x80) && (b < bmax))
+	  {
+	     if (wch < 0) wch = -wch;
+	     *b++ = (unsigned char)(wch);
+	     continue;
+	  }
+
+	dlen = SLUTF8_MAX_MBLEN;
+	if (b + dlen >= bmax)
+	  {
+	     unsigned char *newbuf;
+
+	     dlen = 6;
+	     if (NULL == (newbuf = (unsigned char *)SLrealloc ((char *)buf, buflen+dlen+1)))
+	       {
+		  SLfree ((char *)buf);
+		  SLang_free_array (at);
+		  return;
+	       }
+	     b = newbuf + (b-buf);
+	     buf = newbuf;
+	     buflen += dlen;
+	     bmax = b + buflen;
+	  }
+	b = SLutf8_encode (wch, b, SLUTF8_MAX_MBLEN);
+     }
+
+   if (buf + buflen != b)
+     {
+	unsigned char *newbuf;
+	buflen = b - buf;
+	newbuf = (unsigned char *)SLrealloc((char *)buf, buflen+1);
+	if (newbuf == NULL)
+	  {
+	     SLfree ((char *)buf);
+	     SLang_free_array (at);
+	     return;
+	  }
+	b = newbuf + (b - buf);
+	buf = newbuf;
+     }
+   *b = 0;
+
+   (void) SLang_push_malloced_string ((char *)buf);   /* frees it, even upon error */
+   SLang_free_array (at);
+}
+
 static SLang_Intrin_Fun_Type Strops_Table [] = /*{{{*/
 {
    MAKE_INTRINSIC_I("create_delimited_string",  create_delimited_string_cmd, SLANG_VOID_TYPE),
@@ -2951,7 +3099,8 @@ static SLang_Intrin_Fun_Type Strops_Table [] = /*{{{*/
    MAKE_INTRINSIC_S("strtrans", strtrans_vintrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("str_delete_chars", str_delete_chars_vintrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("glob_to_regexp", glob_to_regexp, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_2("count_char_occurances", count_char_occurances, SLANG_UINT_TYPE, SLANG_STRING_TYPE, SLANG_WCHAR_TYPE),
+   MAKE_INTRINSIC_2("count_char_occurrences", count_char_occurrences, SLANG_UINT_TYPE, SLANG_STRING_TYPE, SLANG_WCHAR_TYPE),
+   MAKE_INTRINSIC_2("count_char_occurances", count_char_occurrences, SLANG_UINT_TYPE, SLANG_STRING_TYPE, SLANG_WCHAR_TYPE),
    MAKE_INTRINSIC_0("strskipbytes", skip_bytes_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("islower", islower_intrin, SLANG_CHAR_TYPE),
    MAKE_INTRINSIC_0("isupper", isupper_intrin, SLANG_CHAR_TYPE),
@@ -2968,6 +3117,8 @@ static SLang_Intrin_Fun_Type Strops_Table [] = /*{{{*/
    MAKE_INTRINSIC_0("isascii", isascii_intrin, SLANG_CHAR_TYPE),
    MAKE_INTRINSIC_0("strskipchar", strskipchar_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("strbskipchar", strbskipchar_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("string_to_wchars", string_to_wchars, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("wchars_to_string", wchars_to_string, SLANG_VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
